"插件市场客户端"

import asyncio
import os
import platform
import shlex
import time
import traceback

import requests
import json

from . import urlmethod
from .cfg import Cfg
from .color_print import Print
from .constants import (
    PLUGIN_MARKET_SOURCE_OFFICIAL,
    TOOLDELTA_CLASSIC_PLUGIN,
    TOOLDELTA_INJECTED_PLUGIN,
    TOOLDELTA_PLUGIN_CFG_DIR,
    TOOLDELTA_PLUGIN_DATA_DIR,
)
from .plugin_load import PluginRegData, PluginsPackage
from .utils import Utils

if platform.system().lower() == "windows":
    CLS_CMD = "cls"
else:
    CLS_CMD = "clear"


def clear_screen() -> None:
    "清屏"
    os.system(shlex.quote(CLS_CMD))


def url_join(*urls: str) -> str:
    """连接 URL

    Returns:
        str: 连接后的 URL
    """
    return "/".join(urls)


def get_json_from_url(url: str) -> dict:
    """从 URL 获取 JSON 数据

    Args:
        url (str): URL

    Raises:
        requests.RequestException: Url 请求失败
        requests.RequestException: 服务器返回了不正确的答复

    Returns:
        dict: JSON 数据
    """
    try:
        resp = requests.get(url, timeout=5)
        resp.raise_for_status()
        if resp.status_code != 200:
            raise ValueError(f"EXC {resp.status_code}")
        jsons = resp.json()
        if "code" in jsons.keys() and jsons["code"] != 200:
            raise requests.RequestException(f"服务器传回异常 (500): {jsons['message']}")
        return jsons
    except requests.RequestException as exc:
        raise requests.RequestException(
            f"URL 请求失败: {url} \n§6(看起来您要更改配置文件中的链接)"
        ) from exc
    except json.JSONDecodeError as exc:
        raise requests.RequestException(
            f"服务器返回了不正确的答复：{resp.text}"
        ) from exc


class PluginMarket:
    "插件市场类"

    def __init__(self):
        self._plugin_id_name_map: dict | None = None
        try:
            self.plugins_download_url = Cfg().get_cfg(
                "ToolDelta基本配置.json", {"插件市场源": str}
            )["插件市场源"]
        except Exception:
            self.plugins_download_url = PLUGIN_MARKET_SOURCE_OFFICIAL

    def enter_plugin_market(self, source_url: str | None = None, in_game=False) -> None:
        """
        进入插件市场

        Args:
            source_url (str | None, optional): 插件市场源
            in_game (bool, optional): 是否在游戏内调用的插件市场命令
        """
        Print.clean_print("§6正在连接到插件市场..")
        CONTENT_LENGTH = 12

        def display_plugins_and_packages(start_index: int, total_pages: int):
            """
            显示插件列表

            Args:
                start_index (int): 起始索引
                total_pages (int): 总页数
            """
            clear_screen()
            Print.clean_print(
                f"{market_datas['SourceName']}: {market_datas['Greetings']}"
            )
            for i in range(start_index, min(start_index + CONTENT_LENGTH, all_indexes)):
                if show_list[i][0].startswith("[pkg]"):
                    pkg_name = show_list[i][0]
                    Print.clean_print(f" {i + 1}. §c[整合包]§e{pkg_name[5:]}")
                else:
                    plugin_id = show_list[i][0]
                    plugin_name = plugin_ids_map[plugin_id]
                    plugin_basic_datas = show_list[i][1]
                    plugin_type = {"classic": "类式", "injected": "注入式"}.get(
                        plugin_basic_datas.get("plugin-type", "unknown"),
                        plugin_basic_datas.get("plugin-type", "unknown"),
                    )
                    Print.clean_print(
                        f" {i + 1}. §e{plugin_name} §av{plugin_basic_datas['version']} "
                        f"§b@{plugin_basic_datas['author']} §d{plugin_type}插件"
                    )
            Print.clean_print(
                f"§f第§a{start_index // CONTENT_LENGTH + 1}§f/§a{total_pages}§f页, 输入§b+§f/§b-§f翻页"
            )
            Print.clean_print("§f输入插件序号选中插件并查看其下载页")

        def handle_plugin_selection(plugin_data: PluginRegData):
            """处理插件选择

            Args:
                plugin_data (dict): 插件数据

            Returns:
                bool: 是否退出
            """
            ok, pres = self.skim_plugin(plugin_data)
            if ok:
                Print.clean_print("可以输入 §breload§r 使这个插件生效哦")
                return (
                    input(
                        Print.clean_fmt("输入 §cq §r退出, 其他则返回插件市场")
                    ).lower()
                    == "q"
                )
            else:
                Print.clean_print("已取消。")
                time.sleep(1)
            return False

        def handle_package_selection(pack: PluginsPackage):
            ok = self.skim_package(pack)
            if ok:
                Print.clean_print("可以输入 §breload§r 使这个整合包生效哦")
                return (
                    input(
                        Print.clean_fmt("§f输入 §cq §f退出, 其他则返回插件市场")
                    ).lower()
                    == "q"
                )
            else:
                Print.clean_print("已取消。")
                time.sleep(1)
            return False

        try:
            market_datas = self.get_market_datas()
            plugin_ids_map = self.get_plugin_id_name_map()
            show_list = [
                (i, j) if i.startswith("[pkg]") else ("[pkg]" + i, j)
                for i, j in market_datas["Packages"].items()
            ] + list(market_datas["MarketPlugins"].items())
            all_indexes = len(show_list)
            now_index = 0
            sum_pages = (all_indexes - 1) // CONTENT_LENGTH + 1
            last_operation = ""

            while True:
                display_plugins_and_packages(now_index, sum_pages)

                last_operation = (
                    input(
                        Print.clean_fmt(
                            "§f回车键继续上次操作, §bq§f 退出，请输入: "
                        )
                    )
                    or last_operation
                )
                last_operation = last_operation.lower().strip()

                # 翻页操作
                if last_operation in ["+", "-"]:
                    now_index = max(
                        0,
                        min(
                            now_index + (CONTENT_LENGTH if last_operation == "+" else -CONTENT_LENGTH),
                            all_indexes - 1,
                        ),
                    )
                elif last_operation == "q":  # 退出操作
                    break
                else:
                    res = Utils.try_int(last_operation)
                    if res and 1 <= res <= all_indexes:  # 插件选择操作
                        result = show_list[res - 1]
                        if not result[0].startswith("[pkg]"):
                            # 这是插件
                            plugin_data = self.get_plugin_data_from_market(result[0])
                            if handle_plugin_selection(plugin_data):
                                break
                        else:
                            # 这是整合包
                            package_data = self.get_package_from_market(result[0])
                            if handle_package_selection(package_data):
                                break
                    else:  # 超出序号范围
                        Print.clean_print("§c超出序号范围")

        except (KeyError, requests.RequestException) as err:
            Print.clean_print(f"§c获取插件市场插件出现问题({err.__class__.__name__}): {err}")
            input(Print.clean_fmt("§6按回车键继续.."))
        except Exception:
            Print.clean_print("§c获取插件市场插件出现问题, 报错如下:")
            Print.clean_print("§c" + traceback.format_exc().replace("\n", "\n§c"))
            input(Print.clean_fmt("§6按回车键继续.."))
        finally:
            Print.clean_print("§a已从插件市场返回 ToolDelta 控制台。")

    def get_market_datas(self) -> dict:
        """
        从插件市场的 market_tree.json 获取数据

        Args:
            source_url (str | None, optional): 插件市场源，默认为 None, 有则替换整体插件市场源

        Returns:
            dict: 插件市场数据
        """
        market_datas = get_json_from_url(
            url_join(self.plugins_download_url, "market_tree.json")
        )
        return market_datas

    def get_market_filetree(self) -> dict:
        """
        从插件市场的 market_tree.json 获取数据

        Args:
            source_url (str | None, optional): 插件市场源，默认为 None, 有则替换整体插件市场源

        Returns:
            dict: 插件市场数据
        """
        if not hasattr(self, "_market_filetree"):
            self._market_filetree = get_json_from_url(
                url_join(self.plugins_download_url, "directory.json")
            )
        return self._market_filetree

    def get_plugin_data_from_market(self, plugin_id: str) -> PluginRegData:
        """从插件市场获取单个插件数据

        Args:
            plugin_id (str): 插件 ID

        Raises:
            KeyError: 无法通过 ID 查找插件

        Returns:
            PluginRegData: 插件注册数据
        """
        plugin_name = self.get_plugin_id_name_map().get(plugin_id)
        if plugin_name is None:
            raise requests.RequestException(
                f"无法通过 ID: {plugin_id} 查找插件, 你可能需要反馈此问题至开发者"
            )
        data_url = self.plugins_download_url + "/" + plugin_name + "/datas.json"
        datas = get_json_from_url(data_url)
        return PluginRegData(plugin_name, datas)

    def get_package_from_market(self, name: str) -> PluginsPackage:
        target_data_url = url_join(self.plugins_download_url, name, "datas.json")
        content = requests.get(target_data_url).json()
        return PluginsPackage(name, content)

    def skim_plugin(
        self, plugin_data: PluginRegData
    ) -> tuple[bool, list[PluginRegData]]:
        """
        选中插件进行介绍与操作

        Args:
            plugin_data (PluginRegData): 插件注册数据

        Returns:
            tuple[bool, list[PluginRegData]]: 是否下载，下载的插件列表
        """
        pre_plugins_str = (
            ", ".join([f"{k}§7v{v}" for k, v in plugin_data.pre_plugins.items()])
            or "无"
        )
        clear_screen()
        Print.clean_print(f"{plugin_data.name} v{plugin_data.version_str}")
        Print.clean_print(
            f"作者: §f{plugin_data.author}§7, 版本: §f{plugin_data.version_str} §b{plugin_data.plugin_type_str}"
        )
        Print.clean_print(f"前置插件：§f{pre_plugins_str}")
        Print.clean_print(f"介绍：{plugin_data.description}")
        Print.clean_print("")
        res = (
            input(Print.clean_fmt("§f下载 = §aY§f, 取消 = §cN§f, 请输入："))
            .lower()
            .strip()
        )
        if res == "y":
            Print.clean_print(f"§6正在下载插件：§f{plugin_data.name}", end="\r")
            pres = self.download_plugin(plugin_data)
            pres.reverse()
            return True, pres
        return False, []

    def skim_package(self, pack: PluginsPackage) -> bool:
        """
        选中整合包进行介绍与操作

        Args:
            pack (PluginsPackage): 整合包数据类

        Returns:
            bool: 是否下载安装
        """
        clear_screen()
        inc_plugins_name: list[str] = []
        for pid in pack.plugin_ids:
            if pname := self.get_plugin_id_name_map().get(pid):
                inc_plugins_name.append(pname)
            else:
                Print.clean_print(f"§c无法通过ID {pid} 查找插件, 有可能是插件市场出错")
                return True
        Print.clean_print(f"§f整合包 §b{pack.name[5:]} §7(v{pack.version})§r:")
        Print.clean_print(f"作者: §b{pack.author}")
        Print.clean_print("介绍: §f" + pack.description.replace("\n", "\n      "))
        Print.clean_print("§d包含的插件的列表:")
        for pname in inc_plugins_name:
            Print.clean_print(f" §7- §r{pname}")
        # 显示其他文件数量
        ftree = self.get_market_filetree()
        dirdata = ftree.get(pack.name)
        if dirdata is None:
            raise ValueError(f"插件市场内不存在整合包 {pack.name}")
        plugin_config_files = ftree[url_join(pack.name, "插件配置文件")]
        # 计算插件配置文件数量
        config_files_num = len(plugin_config_files)
        # 计算插件数据文件数量
        data_files_num = 0
        for folder_path, inc_files in ftree.items():
            spliter = folder_path.split("/")
            if len(spliter) < 2:
                continue
            if spliter[0] == pack.name and spliter[1] == "插件数据文件":
                data_files_num += len(inc_files)
        Print.clean_print(f"§2并包含§r{config_files_num}§2个插件配置文件, §r{data_files_num}§2个插件数据文件")
        if (
            input(Print.clean_fmt("§f下载安装 = §aY§f, 取消 = §cN§f, 请输入："))
            .lower()
            .strip()
        ) == "y":
            self.user_get_plugin_package(pack)
            return True
        else:
            return False

    def get_plugin_id_name_map(self) -> dict[str, str]:
        """
        获取插件 ID 与插件名的映射

        Returns:
            dict: 插件 ID 与插件名的映射
        """
        if self._plugin_id_name_map is None:
            try:
                res = requests.get(
                    self.plugins_download_url + "/plugin_ids_map.json", timeout=5
                )
                res.raise_for_status()
                res1: dict = json.loads(res.text)
            except Exception as err:
                Print.clean_print(
                    f"§c从 {self.plugins_download_url} 获取插件信息遇到问题: {err}"
                )
                raise SystemExit
            self._plugin_id_name_map = res1
            return res1
        else:
            return self._plugin_id_name_map

    def get_download_list(self, plugin_data: PluginRegData) -> dict[str, PluginRegData]:
        """
        获取一个插件的下载列表

        Args:
            plugin_data (PluginRegData): 插件注册数据

        Returns:
            list[str]: 下载列表
        """
        download_paths = {}
        stack = [plugin_data]
        while stack:
            current_plugin = stack.pop()
            download_paths[current_plugin.name] = current_plugin
            for plugin_id in current_plugin.pre_plugins:
                plugin_datas = self.get_plugin_data_from_market(plugin_id)
                stack.append(plugin_datas)
        return download_paths

    def download_plugin(
        self, plugin_data: PluginRegData, with_pres=True, is_enabled=True
    ) -> list[PluginRegData]:
        """
        下载插件
        注意：只能传入由 `get_plugin_data_from_market()` 生成的数据

        Args:
            plugin_data (PluginRegData): 插件注册数据
            with_pres (bool): 是否一同下载前置插件
            is_enabled (bool): 下载的插件是否要自动禁用

        Raises:
            ValueError: 未知插件类型

        Returns:
            list[PluginRegData]: 本插件和其前置插件的注册数据列表
        """
        # 打印正在获取的插件下载树
        Print.clean_print(
            f"§6正在获取 §f{plugin_data.name} §6插件的下载任务清单.." + " " * 15,
            end="\r",
        )
        plugin_list = self.get_download_list(plugin_data)
        plugin_filepaths_dict = {}
        for k, v in plugin_list.items():
            plugin_filepaths_dict[k] = self.find_dirs(v)
        Print.clean_print(f"§a已获取插件下载清单 §f{plugin_data.name}§a" + " " * 15)
        plugins_url2dst_solve: list[tuple[str, str]] = []
        for plugin_name, pluginInfo in plugin_list.items():
            match pluginInfo.plugin_type:
                case "classic":
                    plugintype_path = os.path.join("插件文件", TOOLDELTA_CLASSIC_PLUGIN)
                case "injected":
                    plugintype_path = os.path.join(
                        "插件文件", TOOLDELTA_INJECTED_PLUGIN
                    )
                case _:
                    raise ValueError(
                        f"未知插件类型：{pluginInfo.plugin_type}, 你可能需要通知 ToolDelta 项目开发组解决"
                    )
            for path in plugin_filepaths_dict[plugin_name]:
                plugins_url2dst_solve.append(
                    (
                        url_join(self.plugins_download_url, path),
                        os.path.join(plugintype_path, path),
                    )
                )
        Print.clean_print(
            f"§bTD下载管理器: §7需要下载 §c{len(plugins_url2dst_solve)} §7个文件"
        )
        asyncio.run(urlmethod.download_file_urls(plugins_url2dst_solve))
        Print.clean_print("§a• 插件安装已完成")
        return list(plugin_list.values())

    def user_get_plugin_package(self, pack: PluginsPackage):
        Print.clean_print("§6获取插件数据中...", end="\r")
        find_plugins = Utils.thread_gather(
            [(self.get_plugin_data_from_market, (i,)) for i in pack.plugin_ids]
        )
        ftree = self.get_market_filetree()
        dirdata = ftree.get(pack.name)
        if dirdata is None:
            raise ValueError(f"插件市场内不存在整合包 {pack.name}")
        plugin_config_files = ftree[url_join(pack.name, "插件配置文件")]
        download_url_dirs: list[tuple[str, str]] = []
        # 插件配置文件
        mydir = TOOLDELTA_PLUGIN_CFG_DIR
        for cfgfile in plugin_config_files:
            f_url = url_join(
                self.plugins_download_url,
                pack.name,
                "插件配置文件",
                cfgfile,
            )
            f_local = os.path.join(mydir, cfgfile)
            if os.path.isfile(f_local) and (
                input(
                    Print.clean_fmt(
                        f"§6配置文件 §r{cfgfile}§6 已存在, 是否替换§r(§a[默认]y§r/§cn§r)§6: "
                    )
                )
                .strip()
                .lower()
                != "n"
            ):
                download_url_dirs.append((f_url, f_local))
        # 插件数据文件
        for folder_path, inc_files in ftree.items():
            spliter = folder_path.split("/")
            if len(spliter) < 2:
                continue
            if spliter[0] == pack.name and spliter[1] == "插件数据文件":
                for inc_file in inc_files:
                    # url: [pkg]pkg_name/插件数据文件/anydir/...
                    # local: 插件数据文件/anydir/...
                    f_url = url_join(self.plugins_download_url, folder_path, inc_file)
                    f_local = os.path.join(
                        TOOLDELTA_PLUGIN_DATA_DIR, *spliter[2:], inc_file
                    )
                    if not os.path.isfile(f_local) or (
                        input(
                            Print.clean_fmt(
                                f"§6数据文件 §r{f_local}§6 已存在, 是否替换§r(§ay§r/§cn[默认]§r)§6: "
                            )
                        )
                        .strip()
                        .lower()
                        == "y"
                    ):
                        download_url_dirs.append((f_url, f_local))
        Print.clean_print(f"§6开始下载整合包 {pack.name.replace('[pkg]', '')}")
        for _, fpath in download_url_dirs:
            # 初始化需要的文件夹路径
            os.makedirs(os.path.dirname(fpath), exist_ok=True)
        asyncio.run(urlmethod.download_file_urls(download_url_dirs))
        # 下载插件主体
        for plugin in find_plugins:
            self.download_plugin(plugin)
        Print.clean_print("整合包下载完成")

    def find_dirs(self, plugin_data: PluginRegData) -> list[str]:
        """
        查找插件目录

        Args:
            plugin_data (PluginRegData): 插件注册数据

        Raises:
            KeyError: 目录结构错误
            Exception: 获取插件市场插件目录结构出现问题

        Returns:
            list[str]: 插件目录列表
        """
        try:
            data = get_json_from_url(
                url_join(self.plugins_download_url, "directory.json")
            )
            data_list = []
            for folder, files in data.items():
                if plugin_data.name == folder.split(r"/")[0]:
                    # 展开
                    for file in files:
                        data_list.append(folder + r"/" + file)
            return data_list
        except KeyError as err:
            Print.clean_print(
                f"§c获取插件市场插件目录结构出现问题：无法找到 {err}, 有可能是未来得及更新目录"
            )
            raise KeyError(f"无法找到 {err}, 有可能是未来得及更新目录") from err
        except Exception as err:
            Print.clean_print(f"§c获取插件市场插件目录结构出现问题：{err}")
            raise KeyError(f"获取插件市场插件目录结构出现问题：{err}") from err

    def get_latest_plugin_version(self, plugin_id: str) -> str:
        """获取最新插件版本

        Args:
            plugin_id (str): 插件 ID

        Raises:
            KeyError: 无法通过 ID 获取最新插件版本

        Returns:
            str: 最新插件版本
        """
        result = get_json_from_url(
            url_join(self.plugins_download_url, "latest_versions.json")
        ).get(plugin_id)
        if isinstance(result, str):
            return result
        raise KeyError(f"无法通过 ID: {plugin_id} 获取最新插件版本")


market = PluginMarket()



"ToolDelta 类式插件"

import importlib
import os
import sys
import traceback
from typing import TYPE_CHECKING, TypeVar
from collections.abc import Callable

from tooldelta.cfg import Cfg
from tooldelta.color_print import Print
from tooldelta.utils import Utils
from tooldelta.constants import (
    TOOLDELTA_PLUGIN_DIR,
    TOOLDELTA_CLASSIC_PLUGIN,
    TOOLDELTA_PLUGIN_DATA_DIR,
    PacketIDS,
)
from .. import (
    NotValidPluginError,
    PluginAPINotFoundError,
    PluginAPIVersionError,
    plugin_is_enabled,
)

if TYPE_CHECKING:
    # 类型注释
    from tooldelta import ToolDelta
    from tooldelta.plugin_load.PluginGroup import PluginGroup

_ON_ERROR_CB = Callable[[str, Exception, str], None]
__caches__ = {"plugin": None, "api_name": "", "frame": None}
plugins_funcs: dict[str, list[tuple[str, Callable]]] = {
    "on_def": [],
    "on_inject": [],
    "on_player_prejoin": [],
    "on_player_join": [],
    "on_player_message": [],
    "on_player_death": [],
    "on_player_leave": [],
    "on_command": [],
    "on_frame_exit": [],
    "on_reload": [],
}
packet_funcs: dict[int, list[Callable]] = {}
broadcast_evts_listener: dict[str, list[Callable]] = {}
loaded_plugin_modules = []


class Plugin:
    "插件主类"

    name: str = ""
    "插件名"
    version = (0, 0, 1)
    "插件版本号"
    author = "?"
    "作者名"
    description = "..."
    "简介"

    __path_created__ = False

    def __init__(self, frame: "ToolDelta"):
        self.frame = frame
        self.game_ctrl = frame.get_game_control()

    @property
    def data_path(self) -> str:
        "该插件的数据文件夹路径 (调用时直接创建数据文件夹)"
        path = os.path.join(TOOLDELTA_PLUGIN_DATA_DIR, self.name)
        if not self.__path_created__:
            os.makedirs(path, exist_ok=True)
            self.__path_created__ = True
        return path

    def make_data_path(self):
        os.makedirs(os.path.join(TOOLDELTA_PLUGIN_DATA_DIR, self.name), exist_ok=True)
        self.__path_created__ = True

    def print(self, msg: str):
        Print.print_inf(f"{self.name}: {msg}")

    def format_data_path(self, *paths: str):
        return os.path.join(self.data_path, *paths)


_PLUGIN_CLS_TYPE = TypeVar("_PLUGIN_CLS_TYPE")


def add_plugin(plugin: type[_PLUGIN_CLS_TYPE]) -> type[_PLUGIN_CLS_TYPE]:
    try:
        if not Plugin.__subclasscheck__(plugin):
            raise NotValidPluginError(f"插件主类必须继承 Plugin 类 而不是 {plugin}")
    except TypeError as exc:
        raise NotValidPluginError(
            f"插件主类必须继承 Plugin 类 而不是 {plugin.__class__}"
        ) from exc
    if __caches__["plugin"] is not None:
        raise NotValidPluginError("调用了多次 @add_plugin")
    if __caches__["frame"] is None:
        Print.clean_print("§d正在以直接运行模式运行插件..")
        return plugin
    plugin_ins = plugin(__caches__["frame"])  # type: ignore
    __caches__["plugin"] = plugin_ins
    return plugin


def add_plugin_as_api(apiName: str):
    def _add_plugin_2_api(api_plugin: type[_PLUGIN_CLS_TYPE]) -> type[_PLUGIN_CLS_TYPE]:
        if not Plugin.__subclasscheck__(api_plugin):
            raise NotValidPluginError("API 插件主类必须继承 Plugin 类")
        if __caches__["plugin"] is not None:
            raise NotValidPluginError("调用了多次 @add_plugin")
        if __caches__["frame"] is None:
            Print.clean_print("§d正在以直接运行模式运行插件..")
            return api_plugin
        plugin_ins = api_plugin(__caches__["frame"])  # type: ignore
        __caches__["plugin"] = plugin_ins
        __caches__["api_name"] = apiName
        return api_plugin

    return _add_plugin_2_api


# Plugin get and execute


def read_plugins(plugin_grp: "PluginGroup") -> None:
    """
    读取插件

    Args:
        plugin_grp (PluginGroup): 插件组
    """
    PLUGIN_PATH = os.path.join(TOOLDELTA_PLUGIN_DIR, TOOLDELTA_CLASSIC_PLUGIN)
    if PLUGIN_PATH not in sys.path:
        sys.path.append(PLUGIN_PATH)
    broadcast_evts_listener.clear()
    packet_funcs.clear()
    for plugin_dir in os.listdir(PLUGIN_PATH):
        if not plugin_is_enabled(plugin_dir):
            continue
        if os.path.isdir(os.path.join(PLUGIN_PATH, plugin_dir)):
            sys.path.append(os.path.join(PLUGIN_PATH, plugin_dir))
            load_plugin(plugin_grp, plugin_dir)
            if os.path.isfile(
                data_path := os.path.join(
                    "插件文件", TOOLDELTA_CLASSIC_PLUGIN, plugin_dir, "datas.json"
                )
            ):
                plugin_data = Utils.JsonIO.SafeJsonLoad(data_path)
                plugin_grp.loaded_plugin_ids.append(plugin_data["plugin-id"])


def load_plugin(plugin_group: "PluginGroup", plugin_dirname: str) -> None | Plugin:
    """加载插件

    Args:
        plugin_group (PluginGroup): 插件组类
        plugin_dirname (str): 插件目录名
        hot_load (bool, optional): 是否热加载

    Raises:
        ValueError: 插件组未初始化读取
        ValueError: 插件组未绑定框架
        ValueError: 插件主类需要作者名
        ValueError: 插件主类需要作者名
        NotValidPluginError: 插件 不合法
        SystemExit: 插件名字不合法
        SystemExit: 插件配置文件报错
        SystemExit: 插件读取数据失败

    Returns:
        Union[None, Plugin]: 插件实例
    """
    if isinstance(plugin_group, type(None)):
        raise ValueError("插件组未初始化读取")
    if isinstance(plugin_group.linked_frame, type(None)):
        raise ValueError("插件组未绑定框架")
    __caches__["plugin"] = None
    __caches__["api_name"] = ""
    try:
        if os.path.isfile(
            os.path.join(
                "插件文件", TOOLDELTA_CLASSIC_PLUGIN, plugin_dirname, "__init__.py"
            )
        ):
            plugin_module = importlib.import_module(plugin_dirname)
            if plugin_module in loaded_plugin_modules:
                importlib.reload(plugin_module)
                mode_str = "重载"
            else:
                loaded_plugin_modules.append(plugin_module)
                mode_str = "载入"
        else:
            Print.print_war(f"{plugin_dirname} 文件夹 未发现插件文件，跳过加载")
            return None
        plugin_or_none: Plugin | None = __caches__.get("plugin")
        if plugin_or_none is None:
            raise NotValidPluginError(
                "需要调用 1 次 @plugins.add_plugin 以注册插件主类，然而没有调用"
            )
        plugin: Plugin = plugin_or_none
        if plugin.name is None or plugin.name == "":
            raise ValueError(f"插件主类 {plugin.__class__.__name__} 需要作者名")
        plugin_group.plugins.append(plugin)
        _v0, _v1, _v2 = plugin.version

        # 收集事件监听函数
        for evt_name in plugins_funcs.keys():
            if hasattr(plugin, evt_name):
                plugins_funcs.setdefault(evt_name, [])
                plugins_funcs[evt_name].append((plugin.name, getattr(plugin, evt_name)))

        # 收集到了需要监听的数据包
        if plugin_group._cached_packet_cbs != []:
            for pktType, func in plugin_group._cached_packet_cbs:
                ins_func = getattr(plugin, func.__name__)
                if ins_func is None:
                    raise NotValidPluginError("数据包监听不能在主插件类以外定义")
                if pktType not in packet_funcs.keys():
                    packet_funcs[pktType] = []
                packet_funcs[pktType].append(ins_func)

        # 监听全部数据包的监听器
        # 实话实说, 这里写的乱糟糟的
        if allpkt_func := plugin_group._cached_all_packets_listener:
            allpkt_func = getattr(plugin, allpkt_func.__name__)
            if allpkt_func is None:
                raise NotValidPluginError("数据包监听不能在主插件类以外定义")

            def make_cached_func(name, pktID_1):
                def _allpkt_listener(pkt):
                    allpkt_func(pktID_1, pkt)

                _allpkt_listener.__name__ = name
                return _allpkt_listener

            for pktID in range(304):
                if pktID in PacketIDS.__dict__.values():
                    if pktID in (PacketIDS.IDPyRpc, 304):
                        continue

                    if pktID not in packet_funcs.keys():
                        packet_funcs[pktID] = []
                    packet_funcs[pktID].append(
                        make_cached_func(allpkt_func.__name__, pktID)
                    )

        # 收集到了作为API的插件
        if __caches__["api_name"] != "":
            plugin_group.plugins_api[__caches__["api_name"]] = plugin

        # 收集到了广播监听器
        if plugin_group._cached_broadcast_evts != {}:
            for evt, funcs in plugin_group._cached_broadcast_evts.items():
                for func in funcs:
                    ins_func = getattr(plugin, func.__name__)
                    if ins_func is None:
                        raise NotValidPluginError("广播事件监听不能在主插件类以外定义")
                    if broadcast_evts_listener.get(evt) is None:
                        broadcast_evts_listener[evt] = []
                    broadcast_evts_listener[evt].append(ins_func)

        Print.print_suc(
            f"已{mode_str}插件 §f{plugin.name}§b@{_v0}.{_v1}.{_v2} §a作者: §r{plugin.author}"
        )
        plugin_group.normal_plugin_loaded_num += 1
        return plugin
    except NotValidPluginError as err:
        Print.print_err(f"插件 {plugin_dirname} 不合法：{err.args[0]}")
        raise SystemExit from err
    except Cfg.ConfigError as err:
        Print.print_err(f"插件 {plugin_dirname} 配置文件报错：{err}")
        Print.print_err(
            "你也可以直接删除配置文件，重新启动 ToolDelta 以自动生成配置文件"
        )
        raise SystemExit from err
    except Utils.SimpleJsonDataReader.DataReadError as err:
        Print.print_err(f"插件 {plugin_dirname} 读取数据失败：{err}")
    except plugin_group.linked_frame.SystemVersionException as err:
        Print.print_err(f"插件 {plugin_dirname} 需要更高版本的 ToolDelta 加载：{err}")
        raise SystemExit
    except Exception as err:
        Print.print_err(f"加载插件 {plugin_dirname} 出现问题，报错如下：")
        Print.print_err("§c" + traceback.format_exc())
        raise SystemExit from err
    finally:
        plugin_group._cached_broadcast_evts.clear()
        plugin_group._cached_packet_cbs.clear()
        plugin_group._cached_all_packets_listener = None
    return None


# 魔法方法执行部分


def _init_frame(frame: "ToolDelta"):
    __caches__["frame"] = frame


def reload():
    """系统调用, 重置所有处理函数"""
    for v in plugins_funcs.values():
        v.clear()
    for v in packet_funcs.values():
        v.clear()


def execute_def(onerr: _ON_ERROR_CB) -> None:
    """执行插件的二次初始化方法

    Args:
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法。Defaults to NON_FUNC.

    Raises:
        SystemExit: 缺少前置
        SystemExit: 前置版本过低
    """
    try:
        for name, func in plugins_funcs["on_def"]:
            func()
    except PluginAPINotFoundError as err:
        Print.print_err(f"插件 {name} 需要包含该种接口的前置组件：{err.name}")
        raise SystemExit from err
    except PluginAPIVersionError as err:
        Print.print_err(
            f"插件 {name} 需要该前置组件 {err.name} 版本：{err.m_ver}, 但是现有版本过低：{err.n_ver}"
        )
        raise SystemExit from err
    except Exception as err:
        onerr(name, err, traceback.format_exc())
        raise SystemExit


def execute_init(onerr: _ON_ERROR_CB) -> None:
    """执行插件的连接游戏后初始化方法

    Args:
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    for name, func in plugins_funcs["on_inject"]:
        try:
            func()
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_player_prejoin(player, onerr: _ON_ERROR_CB) -> None:
    """执行玩家加入前的方法

    Args:
        player (_type_): 玩家
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    for name, func in plugins_funcs["on_player_prejoin"]:
        try:
            func(player)
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_player_join(player: str, onerr: _ON_ERROR_CB) -> None:
    """执行玩家加入的方法

    Args:
        player (str): 玩家
        onerr (Callable[[str, Exception, str], None], optional): q 插件出错时的处理方法
    """
    for name, func in plugins_funcs["on_player_join"]:
        try:
            func(player)
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_player_message(
    player: str,
    msg: str,
    onerr: _ON_ERROR_CB,
) -> None:
    """执行玩家消息的方法

    Args:
        player (str): 玩家
        msg (str): 消息
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    pat = f"[{player}] "
    if msg.startswith(pat):
        msg = msg.strip(pat)
    for name, func in plugins_funcs["on_player_message"]:
        try:
            func(player, msg)
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_player_leave(player: str, onerr: _ON_ERROR_CB) -> None:
    """执行玩家离开的方法

    Args:
        player (str): 玩家
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    for name, func in plugins_funcs["on_player_leave"]:
        try:
            func(player)
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_player_death(
    player: str,
    killer: str | None,
    msg: str,
    onerr: _ON_ERROR_CB,
):
    """执行玩家死亡的方法

    Args:
        player (str): 玩家
        killer (str | None): 击杀者
        msg (str): 消息
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    for name, func in plugins_funcs["on_player_death"]:
        try:
            func(player, killer, msg)
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_frame_exit(onerr: _ON_ERROR_CB):
    """执行框架退出的方法

    Args:
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    for name, func in plugins_funcs["on_frame_exit"]:
        try:
            func()
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_reloaded(onerr: _ON_ERROR_CB):
    """执行插件重载的方法

    Args:
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    for name, func in plugins_funcs["on_reload"]:
        try:
            func()
        except Exception as err:
            onerr(name, err, traceback.format_exc())


def execute_command(
    name: str,
    msg: str,
    onerr: _ON_ERROR_CB,
) -> None:
    """执行命令 say 的方法

    Args:
        player (str): 玩家
        cmd (str): 命令
        onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
    """
    for plugin_name, func in plugins_funcs["on_command"]:
        try:
            func(plugin_name, msg)
        except Exception as err:
            onerr(plugin_name, err, traceback.format_exc())


def execute_packet_funcs(pktID: int, pkt: dict, onerr: _ON_ERROR_CB) -> bool:
    """处理数据包监听器

    Args:
        pktID (int): 数据包 ID
        pkt (dict): 数据包

    Returns:
        bool: 是否处理成功
    """
    d = packet_funcs.get(pktID)
    if d:
        for func in d:
            try:
                res = func(pkt)
                if res:
                    return True
            except Exception as err:
                onerr("插件方法:" + func.__name__, err, traceback.format_exc())
    return False
